﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TurtleLib
{
    /// <summary>
    /// TurtleSocketのラッパクラスです．
    /// GUIで対話的に使用することを前提としたイベント駆動型構造で接続，切断，受信,応答到着イベントを発生します．
    /// 応答到着イベントを使用して送信コマンドと応答を同期させることができます．
    /// 全てのイベントをUIスレッドで発生させるためにコンストラクタはUIスレッドで呼び出される必要があります．
    /// </summary>
    public class TurtleCommandSynchroniser : TurtleSocket
    {
        /// <summary>
        /// 応答到着イベント．Send()で送信したコマンドの応答が到着したとき発生します．
        /// 第一引数に送信したコマンド，第二引数にその応答が格納されています．
        /// 文字列末尾の改行('\n', '\r')および'>'は除去されます．
        /// </summary>
        public event System.Action<string, string> ResponseArrived;

        private System.Collections.Generic.List<Byte> _receiveBuffer;
        private bool _synchronising;
        private string _lastCommandString;
        private System.Windows.Threading.DispatcherTimer _timeoutTimer;

        /// <summary>
        /// コンストラクタです．
        /// </summary>
        public TurtleCommandSynchroniser()
        {
            _receiveBuffer = new System.Collections.Generic.List<Byte>();
        }

        /// <summary>
        /// 受信文字列から1行取り出します．
        /// </summary>
        private void TokeniseReceivedStream()
        {
            byte[] buff = new byte[256];

            while(true)
            {
                int read = this.Receive(out buff, buff.Length);

                if (read == 0)
                    break;
                
                for (int i = 0; i < read; i++)
                {
                    _receiveBuffer.Add(buff[i]);
                }
            }

            // 頭が>なら削除
            if (_receiveBuffer[0] == '>')
                _receiveBuffer.RemoveAt(0);

            // 次の>を検索して存在しなければまだ処理しない
            int greaterThanSignPos = _receiveBuffer.FindIndex(b => { return b == (Byte)'>'; });

            if (greaterThanSignPos == 0)
                return;

            // 末尾の改行と>を除去します
            string res = System.Text.Encoding.ASCII.GetString(_receiveBuffer.ToArray()).TrimEnd(new Char[] { '>', '\n', '\r' });

            if (ResponseArrived != null)
                ResponseArrived(_lastCommandString, res);

            base.RaiseSocketMessageEvent("応答は\"" + res + "\"です．");

            _receiveBuffer.Clear();

            SocketReceived -= new System.Action(this.TokeniseReceivedStream);
            _synchronising = false;

            _timeoutTimer.Stop();
            _timeoutTimer = null;

        }

        /// <summary>
        /// 引数のバイト列を送信します．コマンドを入力してください．コマンド末尾にはLFが必要です．
        /// </summary>
        /// <param name="ary">送信バイト列．</param>
        /// <example>例えばpcode\nなど</example>
        public new void Send(byte[] ary)
        {
            if (_synchronising)
                return;

            _lastCommandString = System.Text.Encoding.ASCII.GetString(ary).TrimEnd(new Char[] { '\n', '\r' });

            base.RaiseSocketMessageEvent("コマンド\"" + _lastCommandString + "\"を送信しました．応答待機します．");

            _synchronising = true;
            SocketReceived += new System.Action(this.TokeniseReceivedStream);
            base.Send(ary);


            _timeoutTimer = new System.Windows.Threading.DispatcherTimer();
            _timeoutTimer.Interval = new System.TimeSpan(Timeout * 10000);

            _timeoutTimer.Tick += new System.EventHandler((obj, args) =>
            {
                if (_synchronising)
                {
                    base.RaiseSocketMessageEvent("タイムアウト時間内に応答がありませんでした．");
                    this.Close();
                }

                _timeoutTimer.Stop();
                _timeoutTimer = null;

            });

            _timeoutTimer.Start();

        }

        
        /// <summary>
        /// 引数の文字列を送信します．コマンドを入力してください．コマンド末尾にはLFが必要です．
        /// </summary>
        /// <param name="str">送信バイト列．</param>
        /// <example>例えば"pcode\n"など</example>
        public new void Send(string str)
        {
            this.Send(Encoding.ASCII.GetBytes(str));
        }

    }
}
