﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TurtleLib
{
    /// <summary>
    /// バークレーソケットのTCP用簡易ラッパクラスです．
    /// GUIで対話的に使用することを前提としたイベント駆動型構造で接続，切断，受信イベントを発生します．
    /// 全てのイベントをUIスレッドで発生させるためにコンストラクタはUIスレッドで呼び出される必要があります．
    /// </summary>
    public class TurtleSocket
    {
        private System.Collections.Generic.List<Byte> _receivedBuffer;
        private System.Collections.Generic.List<Byte> _sendBuffer;
        private System.Net.Sockets.Socket _socket;
        private System.Threading.Tasks.TaskScheduler _uiSyncContext;

        /// <summary>
        /// ソケットがブロックする際のタイムアウトを設定/取得します．
        /// 動作の遅いコンピュータではこの値が小さすぎると正しく動作しないことがあります．単位ms
        /// </summary>
        public int Timeout { get; set; }

        /// <summary>
        /// ソケットが自身の状態を文字列で通知するときに発生します．
        /// </summary>
        public event System.Action<string> SocketMessageRaised;

        /// <summary>
        /// ソケットがホストに接続したとき発生します．
        /// </summary>
        public event System.Action SocketConnected;

        /// <summary>
        /// ソケットがホストと切断したとき発生します．
        /// </summary>
        public event System.Action SocketClosed;

        /// <summary>
        /// ホストから新しいバイト列を受け取ったとき発生します．
        /// </summary>
        public event System.Action SocketReceived;


        /// <summary>
        /// SocketMessageRaisedイベントを発生させます．
        /// </summary>
        /// <param name="msg">SocketMessageRaisedで送出するメッセージ</param>
        protected void RaiseSocketMessageEvent(String msg)
        {
            if (SocketMessageRaised != null)
                SocketMessageRaised(msg);
        }


        /// <summary>
        /// コンストラクタ．
        /// </summary>
        public TurtleSocket()
        {
            Timeout = 5000;
            _receivedBuffer = new System.Collections.Generic.List<Byte>();
            _sendBuffer = new System.Collections.Generic.List<Byte>();

            var uiSyncContext = System.Threading.Tasks.TaskScheduler.FromCurrentSynchronizationContext();
        }

        /// <summary>
        /// デストラクタです．
        /// </summary>
        ~TurtleSocket()
        {
            if (_socket != null)
                this.Close();

            _uiSyncContext = null;
        }


        /// <summary>
        /// 指定されたホストに接続を試みます．
        /// </summary>
        /// <param name="host">ホスト名</param>
        /// <param name="port">ポート番号</param>
        public void Connect(string host, int port)
        {
            if (_socket != null)
                return;

            _socket = new System.Net.Sockets.Socket(
                System.Net.Sockets.AddressFamily.InterNetwork,
                System.Net.Sockets.SocketType.Stream,
                System.Net.Sockets.ProtocolType.Tcp);

            _socket.ReceiveTimeout = -1;    // 無期限の受信タイムアウト

            _uiSyncContext = System.Threading.Tasks.TaskScheduler.FromCurrentSynchronizationContext();   // UIスレッドで処理するためのTaskScheduler

            System.Threading.Tasks.Task.Factory.StartNew(() =>
            {
                return System.Threading.Tasks.Task.Factory.StartNew(() =>
                {
                    _socket.Connect(host, port);
                }).Wait(Timeout);

            }).ContinueWith(prevTask =>
            {   
                // この継続タスクはUIスレッドで実行されます．
                try
                {
                    if (!prevTask.Result)
                        throw new System.TimeoutException();
                    else
                    {
                        if(SocketConnected!=null)
                            SocketConnected();
                        
                        RaiseSocketMessageEvent(host + ":" + port.ToString() + "と接続しました．");
                        return true;
                    }
                }
                catch (System.Exception e)
                {
                    this.Close();

                    while (e != null)
                    {
                        RaiseSocketMessageEvent(e.Message);
                        e = e.InnerException;
                    }

                    return false;
                }
            }, _uiSyncContext).ContinueWith(prevTask =>
            {
                // この継続タスクはワーカスレッドで実行されます．
                if (prevTask.Result)
                {
                    ReceiveLoop();
                    SendLoop();
                }
            });

        }

        /// <summary>
        /// 現在の接続を閉じます．
        /// </summary>
        public void Close()
        {
            System.Threading.Tasks.Task.Factory.StartNew(() =>
            {
                if (_socket == null)
                    return false;

                _socket.Close(50);
                _socket = null;

                return true;

            }).ContinueWith(prevTask =>
            {
                if (prevTask.Result)
                {
                    RaiseSocketMessageEvent("接続は閉じました．");
                    if (SocketClosed != null)
                        SocketClosed();
                }

            }, _uiSyncContext);

        }

        /// <summary>
        /// 受信ループ．このクラス内部で使用します．
        /// </summary>
        private void ReceiveLoop()
        {
            // Close()后このタスクループは終了します．
            if (_socket == null)
                return;

            int availableBytes = 0;

            try
            {
                availableBytes = _socket.Available;
            }
            catch (System.NullReferenceException)
            {
                return;
            }

            
            // 受信データがないときは新しいタスクで再びこの関数を呼び出します．
            if (availableBytes == 0)
            {
                System.Threading.Tasks.Task.Factory.StartNew(() =>
                {
                    System.Threading.Thread.Sleep(66);
                    this.ReceiveLoop();
                });

                return;
            }

            var bytes = new byte[availableBytes];

            System.Threading.Tasks.Task.Factory.StartNew(() =>
            {
                return System.Threading.Tasks.Task.Factory.StartNew(() =>
                {
                    _socket.Receive(bytes);

                }).Wait(Timeout);

            }).ContinueWith(prevTask =>
            {
                // 受信データをバッファに格納します．
                try
                {
                    if (!prevTask.Result)
                        throw new System.TimeoutException();
                    else
                    {
                        lock (_receivedBuffer)
                        {
                            for (int i = 0; i < availableBytes; i++)
                                _receivedBuffer.Add(bytes[i]);
                        }

                        if (_socket.Available == 0)
                            System.Threading.Thread.Sleep(33);

                        if (SocketReceived != null)
                            SocketReceived();

                        return true;
                    }
                }
                catch (System.Exception e)
                {
                    // エラー発生時はこのソケットを閉じます．
                    this.Close();

                    while (e != null)
                    {
                        RaiseSocketMessageEvent(e.Message);
                        e = e.InnerException;
                    }

                    return false;
                }

            }, _uiSyncContext).ContinueWith(prevTask =>
            {
                // 新しいタスクで再びこの関数を呼び出します．
                if (prevTask.Result)
                    ReceiveLoop();
            });
        }


        /// <summary>
        /// 送信ループ．このクラス内で使用します．
        /// </summary>
        private void SendLoop()
        {
            if (_socket == null)
                return;

            if (_sendBuffer.Count == 0)
            {
                System.Threading.Tasks.Task.Factory.StartNew(() =>
                {
                    System.Threading.Thread.Sleep(66);
                    this.SendLoop();
                });

                return;
            }


            System.Threading.Tasks.Task.Factory.StartNew(() =>
            {
                byte[] bytes;

                lock (_sendBuffer)
                {
                    bytes = _sendBuffer.ToArray();
                    _sendBuffer.Clear();
                }

                return System.Threading.Tasks.Task.Factory.StartNew(() =>
                {
                    _socket.Send(bytes);

                }).Wait(Timeout);

            }).ContinueWith(prevTask =>
            {
                try
                {
                    if (!prevTask.Result)
                        throw new System.TimeoutException();
                    else
                    {
                        if (_sendBuffer.Count == 0)
                            System.Threading.Thread.Sleep(33);

                        return true;
                    }
                }
                catch (System.Exception e)
                {
                    this.Close();

                    while (e != null)
                    {
                        RaiseSocketMessageEvent(e.Message);
                        e = e.InnerException;
                    }

                    return false;
                }


            }, _uiSyncContext).ContinueWith(prevTask =>
            {
                if (prevTask.Result)
                    SendLoop();
            });

        }

        /// <summary>
        /// 引数のバイト列を現在の接続先に送信します．
        /// </summary>
        /// <param name="ary">送信バイト列</param>
        public void Send(byte[] ary)
        {
            if (_socket == null)
                return;

            lock (_sendBuffer)
            {
                foreach (byte b in ary)
                    _sendBuffer.Add(b);
            }
        }

        /// <summary>
        /// 引数のバイト列を現在の接続先に送信します．
        /// </summary>
        /// <param name="str">送信文字列</param>
        public void Send(string str)
        {
            this.Send(Encoding.ASCII.GetBytes(str));
        }

        /// <summary>
        /// 受信したバイト列があれば取得します．
        /// </summary>
        /// <param name="buffer">バイト列の格納先</param>
        /// <param name="bytesToRead">取得するバイト数</param>
        /// <returns>実際に取得したバイト数</returns>
        public int Receive(out byte[] buffer, int bytesToRead)
        {
            bytesToRead = Math.Min(_receivedBuffer.Count, bytesToRead);
            buffer = new byte[bytesToRead];

            lock (_receivedBuffer)
            {
                for (int i = 0; i < bytesToRead; ++i)
                    buffer[i] = _receivedBuffer[i];

                _receivedBuffer.RemoveRange(0, bytesToRead);
            }

            return bytesToRead;
        }



    }
}
